@@ -1,18 +1,5 @@
 '''
 Created on 2020-10-29
-
-@author: wf
-
- script to read push wikipages from one MediaWiki to another
-  
-  use: python3 wikipush.py --source [sourceId] --target [targetId] --pages [pageTitleList]
-  @param sourceId - id of the source wiki
-  @param targetId - id of the target wiki
-  @param pages - pageList to transfer
-  
-  example: 
-      python3 wikipush -s wikipedia-de -t test Berlin
-
   @author:     wf
   @copyright:  Wolfgang Fahl. All rights reserved.
 
@@ -22,6 +9,8 @@ from mwclient.image import Image
 from wikibot.smw import SMWClient
 #from difflib import Differ
 import difflib
+import datetime
+from git import Repo
 import os
 import re
 from pathlib import Path
@@ -35,25 +24,29 @@ class WikiPush(object):
     '''
     differ=None
     
-    def __init__(self, fromWikiId, toWikId,login=False,verbose=True,debug=False):
+    def __init__(self, fromWikiId, toWikiId=None,login=False,verbose=True,debug=False):
         '''
         Constructor
         '''
         self.verbose=verbose
         self.debug=debug
+        self.fromWiki=None
+        self.toWiki=None
         
         self.fromWikiId=fromWikiId
         if self.fromWikiId is not None:
             self.fromWiki=WikiClient.ofWikiId(fromWikiId)
-        self.toWikiId=toWikId
-        self.toWiki=WikiClient.ofWikiId(toWikId)
+        self.toWikiId=toWikiId
+        if self.toWikiId is not None:
+            self.toWiki=WikiClient.ofWikiId(toWikiId)
         if login and self.fromWikiId is not None:
             self.fromWiki.login()
-        self.toWiki.login()
+        if self.toWiki is not None:
+            self.toWiki.login()
         
     def log(self,msg,end='\n'):
         if self.verbose:
-                print (msg,end=end)
+            print (msg,end=end)
      
     def query(self,askQuery,wiki=None,queryField=None):
         '''
@@ -186,6 +179,56 @@ class WikiPush(object):
                 self.uploadImage(file, filename, description, force)
             except Exception as ex:
                 self.log("❌:%s" % str(ex) )
+                
+    def backup(self,pageTitles,backupPath=None,git=False,withImages=False):
+        '''
+        backup the given page titles
+        Args:
+            pageTitles(list): a list of page titles to be downloaded from the fromWiki
+            git(bool): True if git should be used as a version control system
+            withImages(bool): True if the image on a page should also be copied
+        '''
+        if backupPath is None:
+            backupPath=self.getHomePath("wikibackup/%s" % self.fromWikiId)
+        imageBackupPath="%s/images" % backupPath
+        total=len(pageTitles)
+        self.log("downloading %d pages from %s to %s" % (total,self.fromWikiId,backupPath))
+        for i,pageTitle in enumerate(pageTitles):
+            try:
+                self.log("%d/%d (%4.0f%%): downloading %s ..." % (i+1,total,(i+1)/total*100,pageTitle), end='')
+                page=self.fromWiki.getPage(pageTitle)
+                wikiFilePath="%s/%s.wiki" % (backupPath,pageTitle)
+                self.ensureParentDirectoryExists(wikiFilePath)
+                with open (wikiFilePath,"w") as wikiFile:
+                    wikiFile.write(page.text())
+                self.log("✅")
+                if isinstance(page,Image):
+                    self.backupImages([page],imageBackupPath)
+                if withImages:
+                    self.backupImages(page.images(), imageBackupPath)    
+                    
+            except Exception as ex:
+                self.log("❌:%s" % str(ex) )
+        if git:
+            gitPath="%s/.git" % backupPath
+            if not os.path.isdir(gitPath):
+                self.log("initializing git repository ...")
+                repo=Repo.init(backupPath)
+            else:
+                repo=Repo(backupPath)
+            self.log("committing to git repository")
+            repo.git.add(all=True)
+            timestamp=datetime.datetime.now().isoformat()
+            repo.index.commit("auto commit by wikibackup at %s" % timestamp)
+        
+    def backupImages(self,imageList,imageBackupPath):
+        '''
+        '''
+        for image in imageList:
+            try:
+                imagePath,filename=self.downloadImage(image,imageBackupPath);
+            except Exception as ex:
+                self.handleException(ex)
         
     def push(self,pageTitles,force=False,ignore=False,withImages=False):
         '''
@@ -206,14 +249,18 @@ class WikiPush(object):
                 if page.exists:
                     # is this an image?
                     if isinstance(page,Image):
-                        self.pushImages([page], ignore)
+                        self.pushImages([page],ignore=ignore)
                     else:
                         newPage=self.toWiki.getPage(pageTitle)
                         if not newPage.exists or force:
-                            comment="pushed from %s by wikipush" % self.fromWikiId
-                            newPage.edit(page.text(),comment)
-                            self.log("✅")
-                            if withImages:
+                            try:
+                                comment="pushed from %s by wikipush" % self.fromWikiId
+                                newPage.edit(page.text(),comment)
+                                self.log("✅")
+                                pageOk=True
+                            except Exception as ex:
+                                pageOk=self.handleException(ex, ignore)
+                            if withImages and pageOk:
                                 self.pushImages(page.images(),ignore=ignore)
                         else:
                             self.log("👎")
@@ -222,16 +269,30 @@ class WikiPush(object):
             except Exception as ex:
                 self.log("❌:%s" % str(ex) )
                 
+    def ensureParentDirectoryExists(self,filePath):
+        # for pages that have a "/" in the name:
+        directory = os.path.dirname(filePath)
+        self.ensureDirectoryExists(directory)
+        
+    def ensureDirectoryExists(self,directory):
+        Path(directory).mkdir(parents=True, exist_ok=True)
+        
+    def getHomePath(self,localPath):
+        '''
+        get the given home path
+        '''
+        homePath=str(Path.home() / localPath)
+        self.ensureDirectoryExists(homePath)
+        return homePath
+                    
     def getDownloadPath(self):
         '''
         get the download path
         '''
-        downloadPath=str(Path.home() / "Downloads/mediawiki")
-        if not os.path.isdir(downloadPath):
-            os.makedirs(downloadPath)
-        return downloadPath
+        return self.getHomePath("Downloads/mediawiki")
+        
             
-    def pushImages(self,imageList,ignore=False):
+    def pushImages(self,imageList,delim="",ignore=False):
         '''
         push the images in the given image List
         
@@ -241,22 +302,82 @@ class WikiPush(object):
         '''        
         for image in imageList:
             try:
-                filename=image.name.replace("File:","")
-                downloadPath=self.getDownloadPath()
-                imagePath="%s/%s" % (downloadPath,filename)
-                self.log("copying image %s ..." % image.name, end='')
-                with open(imagePath,'wb') as imageFile:
-                    image.download(imageFile)
-                    
+                self.log("%scopying image %s ..." % (delim,image.name), end='')  
+                imagePath,filename=self.downloadImage(image);
                 description=image.imageinfo['comment'] 
-                #imageFile=io.BytesIO(imageContent)
-                self.uploadImage(imagePath,filename,description)
+                self.uploadImage(imagePath,filename,description,ignore)
                 if self.debug:
                     print(image.imageinfo)
             except Exception as ex:
-                self.log("❌:%s" % str(ex)) 
+                self.handleException(ex,ignore)
+                
+    def handleException(self,ex,ignoreExists=False):
+        '''
+        handle the given exception and ignore it if it includes "exists" and ignoreExists is True
+        
+        Args:
+            ex(Exception): the exception to handle
+            ignoreExists(bool): True if "exists" should be ignored
+            
+        Returns:
+            bool: True if the exception was handled as ok False if it was logged as an error
+        '''
+        msg=str(ex)
+        return self.handleWarning(msg,marker="❌",ignoreExists=ignoreExists)
+        
+    def handleAPIWarnings(self,warnings,ignoreExists=False):
+        '''
+        handle API Warnings
+        
+        Args:
+            warnings(list): a list of API warnings
+            ignoreExists(bool): ignore messages that warn about existing content
+        '''
+        msg=""
+        if warnings:
+            for warning in warnings:
+                msg+="%s\n" % str(warning)
+        self.handleWarning(msg,ignoreExists)
+        
+    def handleWarning(self,msg,marker="⚠️",ignoreExists=False):
+        '''
+        handle the given warning and ignore it if it includes "exists" and ignoreExists is True
+        
+        Args:
+            msg(string): the warning to handle
+            marker(string): the marker to use for the message
+            ignoreExists(bool): True if "exists" should be ignored
+            
+        Returns:
+            bool: True if the exception was handled as ok False if it was logged as an error
+        '''
+        #print ("handling warning %s with ignoreExists=%r" % (msg,ignoreExists))
+        if ignoreExists and "exists" in msg:
+            # shorten exact duplicate message
+            if "exact duplicate in msg":
+                msg="exact duplicate"
+            marker="👀"
+        self.log("%s:%s" % (marker,msg))
+        return marker=="👀"
+                
+    def downloadImage(self,image,downloadPath=None):
+        '''
+        download the given image
+        
+        Args:
+            image(image): the image to download
+            downloadPath(str): the path to download to if None getDownloadPath will be used
+        '''
+        filename=image.name.replace("File:","")
+        if downloadPath is None:
+            downloadPath=self.getDownloadPath()
+        imagePath="%s/%s" % (downloadPath,filename)
+        self.ensureParentDirectoryExists(imagePath)
+        with open(imagePath,'wb') as imageFile:
+            image.download(imageFile)
+        return imagePath,filename
                             
-    def uploadImage(self,imagePath,filename,description,ignore=False):
+    def uploadImage(self,imagePath,filename,description,ignoreExists=False):
         '''
         upload an image
         
@@ -264,27 +385,24 @@ class WikiPush(object):
             imagePath(str): the path to the image
             filename(str): the filename to use
             description(str): the description to use
-            ignore(bool): True if it should be ignored if the image exists
+            ignoreExists(bool): True if it should be ignored if the image exists
         '''
         with open(imagePath,'rb') as imageFile:
             warnings=None
-            response=self.toWiki.site.upload(imageFile,filename,description,ignore=ignore)
+            response=self.toWiki.site.upload(imageFile,filename,description,ignoreExists)
             if 'warnings' in response:
                 warnings=response['warnings']
             if 'upload' in response and 'warnings' in response['upload']:
                 warningsDict=response['upload']['warnings']
                 warnings=[]
                 for item in warningsDict.items():
-                    warnings.append(item)
-            if warnings is not None:
-                self.log("❌:%s" % warnings)  
-            else:
-                self.log("✅")
+                    warnings.append(str(item))
+            self.handleAPIWarnings(warnings,ignoreExists)
        
 
-__version__ = 0.1
+__version__ = "0.1.15"
 __date__ = '2020-10-31'
-__updated__ = '2020-10-31'
+__updated__ = '2020-12-05'
 DEBUG=False
 
 def mainNuke(argv=None):
@@ -298,21 +416,21 @@ def mainPush(argv=None):
     
 def mainUpload(argv=None):
     main(argv,mode='wikiupload')
-        
+
+def mainBackup(argv=None):
+    main(argv,mode='wikibackup')
     
 def main(argv=None,mode='wikipush'): # IGNORE:C0111
     '''main program.'''
 
     if argv is None:
-        argv = sys.argv
-    else:
-        sys.argv.extend(argv)   
+        argv = sys.argv[1:]
     
     program_name = mode
     program_version = "v%s" % __version__
     program_build_date = str(__updated__)
     program_version_message = '%%(prog)s %s (%s)' % (program_version, program_build_date)
-    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
+    program_shortdesc = "wikipush"
     user_name="Wolfgang Fahl"
     
     program_license = '''%s
@@ -326,7 +444,6 @@ def main(argv=None,mode='wikipush'): # IGNORE:C0111
   Distributed on an "AS IS" basis without warranties
   or conditions of any kind, either express or implied.
 
-USAGE
 ''' % (program_shortdesc,user_name, str(__date__))
 
     try:
@@ -340,6 +457,11 @@ USAGE
             parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to overwrite existing pages")
             parser.add_argument("-i", "--ignore", dest="ignore", action='store_true', help="ignore upload warnings e.g. duplicate images")
             parser.add_argument("-wi", "--withImages", dest="withImages", action='store_true', help="copy images on the given pages")
+        elif mode=="wikibackup": 
+            parser.add_argument("-g", "--git", dest="git", action='store_true', help="use git for version control")
+            parser.add_argument("-l", "--login", dest="login", action='store_true', help="login to source wiki for access permission")
+            parser.add_argument("-s", "--source", dest="source", help="source wiki id", required=True)
+            parser.add_argument("-wi", "--withImages", dest="withImages", action='store_true', help="copy images on the given pages")
         elif mode=="wikinuke":
             parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to delete pages - default is 'dry' run only listing pages")            
         elif mode=="wikiedit":
@@ -351,18 +473,22 @@ USAGE
             parser.add_argument("--files", nargs='+', help="list of files to be uploaded", required=True)
             parser.add_argument("-f", "--force", dest="force", action='store_true', help="force to (re)upload existing files - default is false")            
             pass
-        if mode in  ["wikipush","wikiedit","wikinuke"]: 
+        if mode in  ["wikipush","wikiedit","wikinuke","wikibackup"]: 
             parser.add_argument("-q", "--query", dest="query", help="select pages with given SMW ask query", required=False)
             parser.add_argument("-qf", "--queryField",dest="queryField",help="query result field which contains page")
             parser.add_argument("-p", "--pages", nargs='+', help="list of page Titles to be pushed", required=False)
         
-        parser.add_argument("-t", "--target", dest="target", help="target wiki id", required=True)    
+        if not mode=="wikibackup":
+            parser.add_argument("-t", "--target", dest="target", help="target wiki id", required=True)    
         # Process arguments
-        args = parser.parse_args()
+        args = parser.parse_args(argv)
         
         if mode=="wikipush":
             wikipush=WikiPush(args.source,args.target,login=args.login)
             queryWiki=wikipush.fromWiki
+        elif mode=="wikibackup":
+            wikipush=WikiPush(args.source,None,login=args.login)    
+            queryWiki=wikipush.fromWiki
         elif mode=="wikiupload":
             wikipush=WikiPush(None,args.target)
         else:
@@ -371,13 +497,18 @@ USAGE
         if mode=="wikiupload":
             wikipush.upload(args.files,args.force)
         else:    
+            pages=None
             if args.pages:
                 pages=args.pages
             elif args.query:
                 pages=wikipush.query(args.query,wiki=queryWiki,queryField=args.queryField)
-            if pages:
+            if pages is None:
+                raise Exception("no pages specified - you might want to use the -p or -q option")
+            else:
                 if mode=="wikipush":
                     wikipush.push(pages,force=args.force,ignore=args.ignore,withImages=args.withImages)
+                elif mode=="wikibackup":
+                    wikipush.backup(pages,git=args.git,withImages=args.withImages)    
                 elif mode=='wikinuke':
                     wikipush.nuke(pages,force=args.force)
                 elif mode=='wikiedit':
@@ -385,7 +516,6 @@ USAGE
                     wikipush.edit(pages,modify=modify,context=args.context,force=args.force)
                 else:
                     raise Exception("undefined wikipush mode %s" % mode)
-      
         
     except KeyboardInterrupt:
         ### handle keyboard interrupt ###
